@@ -7,12 +7,12 @@
  * modify it under the terms of the GNU General Public
  * License as published by the Free Software Foundation; either
  * version 2 of the License, or (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public
  * License along with this program; if not, write to the
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
@@ -27,20 +27,50 @@
  *  - my_free()
  */
 
-#define _LARGEFILE64_SOURCE
-#define _XOPEN_SOURCE 600
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif	/* HAVE_CONFIG_H */
 
+#if HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif	/* HAVE_SYS_TYPES_H */
+
 #include <stdio.h>
+
+#if HAVE_STDLIB_H
 #include <stdlib.h>
+#endif	/* HAVE_STDLIB_H */
+
+#if HAVE_UNISTD_H
 #include <unistd.h>
+#endif	/* HAVE_UNISTD_H */
+
+#if HAVE_FCNTL_H
 #include <fcntl.h>
+#endif	/* HAVE_FCNTL_H */
+
+#if HAVE_STRINGS_H
 #include <strings.h>
-#include <stdarg.h>
+#endif	/* HAVE_STRINGS_H */
+
+#if HAVE_STRING_H
 #include <string.h>
+#endif	/* HAVE_STRING_H */
+
+#if HAVE_SYS_MOUNT_H
 #include <sys/mount.h>
+#endif	/* HAVE_SYS_MOUNT_H */
+
+#if HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif	/* HAVE_SYS_STAT_H */
+
+#if HAVE_SYSLOG_H
+#include <syslog.h>
+#endif	/* HAVE_SYSLOG_H */
+
 #include <signal.h>
+#include <stdarg.h>
 #include <errno.h>
 
 #include "fstab.h"
@@ -51,8 +81,10 @@
 #include "mount_constants.h"
 #include "mount_opts.h"
 #include "mount.nilfs2.h"
+#include "cleaner_exec.h"
 #include "nls.h"
 
+/* mount options */
 int verbose = 0;
 int mount_quiet = 0;
 int readonly = 0;
@@ -61,6 +93,7 @@ static int nomtab = 0;
 static int devro = 0;
 static int fake = 0;
 
+/* global variables */
 extern char *optarg;
 extern int optind;
 
@@ -70,6 +103,12 @@ char *progname = "mount." NILFS2_FS_NAME;
 const char gcpid_opt_fmt[] = PIDOPT_NAME "=%d";
 typedef int gcpid_opt_t;
 
+const char pp_opt_fmt[] = PPOPT_NAME "=%lu";
+typedef unsigned long pp_opt_t;
+
+const char nogc_opt_fmt[] = NOGCOPT_NAME;
+typedef int nogc_opt_t;
+
 struct mount_options {
 	char *fstype;
 	char *opts;
@@ -79,6 +118,19 @@ struct mount_options {
 struct mount_options options;
 
 
+static void nilfs_mount_logger(int priority, const char *fmt, ...)
+{
+	va_list args;
+
+	if ((verbose && priority > LOG_INFO) || priority >= LOG_INFO)
+		return;
+	va_start(args, fmt);
+	fprintf(stderr, "%s: ", progname);
+	vfprintf(stderr, fmt, args);
+	fputs(_("\n"), stderr);
+	va_end(args);
+}
+
 /* Report on a single mount.  */
 static void print_one (const struct my_mntent *me)
 {
@@ -137,11 +189,16 @@ static int device_is_readonly(const char *device, int *ro)
 	return 0;
 }
 
+static void show_version(void)
+{
+	printf("%s (%s %s)\n", progname, PACKAGE, PACKAGE_VERSION);
+}
+
 static void parse_options(int argc, char *argv[], struct mount_options *opts)
 {
-	int c;
+	int c, show_version_only = 0;
 
-	while ((c = getopt(argc, argv, "fvnt:o:rw")) != EOF) {
+	while ((c = getopt(argc, argv, "fvnt:o:rwV")) != EOF) {
 		switch (c) {
 		case 'f':
 			fake++;
@@ -168,11 +225,19 @@ static void parse_options(int argc, char *argv[], struct mount_options *opts)
 		case 'w':
 			readwrite++;
 			break;
+		case 'V':
+			show_version_only++;
+			break;
 		default:
 			break;
 		}
 	}
 
+	if (show_version_only) {
+		show_version();
+		exit(0);
+	}
+
 	parse_opts(opts->opts, &opts->flags, &opts->extra_opts);
 }
 
@@ -202,7 +267,7 @@ static int mounted(const char *spec, const char *node)
 	mc = getmntdirbackward(dir, NULL);
 	while (mc) {
 		if (strcmp(mc->m.mnt_type, fstype) == 0 &&
-		    strcmp(mc->m.mnt_fsname, fsname) == 0) {
+		    (fsname == NULL || strcmp(mc->m.mnt_fsname, fsname) == 0)) {
 			ret = 1;
 			break;
 		}
@@ -224,16 +289,29 @@ static void update_gcpid_opt(char **opts, pid_t newpid)
 	*opts = newopts;
 }
 
-static char *fix_extra_opts_string(const char *extra_opts, pid_t gcpid)
+static char *fix_extra_opts_string(const char *extra_opts, pid_t gcpid,
+				   pp_opt_t protection_period)
 {
 	char buf[256];
 	gcpid_opt_t id;
+	pp_opt_t pp;
+	char *tmpres;
+	char *res;
 
 	buf[0] = '\0';
 	if (gcpid)
 		snprintf(buf, sizeof(buf), gcpid_opt_fmt, (int)gcpid);
 	/* The gcpid option will be removed if gcpid == 0 */
-	return change_opt(extra_opts, gcpid_opt_fmt, &id, buf);
+	tmpres = change_opt(extra_opts, gcpid_opt_fmt, &id, buf);
+
+	buf[0] = '\0';
+	if (protection_period != ULONG_MAX)
+		snprintf(buf, sizeof(buf), pp_opt_fmt, protection_period);
+	/* The pp option will be removed if pp == ULONG_MAX */
+	res = change_opt(tmpres, pp_opt_fmt, &pp, buf);
+
+	my_free(tmpres);
+	return res;
 }
 
 /*
@@ -311,6 +389,8 @@ struct nilfs_mount_info {
 	pid_t gcpid;
 	int type;
 	int mounted;
+	pp_opt_t protperiod;
+	nogc_opt_t nogc;
 };
 
 static int check_mtab(void)
@@ -334,50 +414,61 @@ prepare_mount(struct nilfs_mount_info *mi, const struct mount_options *mo)
 {
 	struct mntentchn *mc;
 	gcpid_opt_t pid;
+	pp_opt_t prot_period;
 	int res = -1;
 
+	if (!(mo->flags & MS_REMOUNT) && mounted(NULL, mi->mntdir)) {
+		error(_("%s: %s is already mounted."), progname, mi->mntdir);
+		goto failed;
+	}
+
 	mi->type = NORMAL_MOUNT;
 	mi->gcpid = 0;
 	mi->optstr = NULL;
 	mi->mounted = mounted(mi->device, mi->mntdir);
+	mi->protperiod = ULONG_MAX;
 
-	if (!(mo->flags & MS_RDONLY)) { /* rw-mount */
-		if ((mc = find_rw_mount(mi->device)) == NULL)
-			goto out;
-		if (!(mo->flags & MS_REMOUNT)) {
-			error(_("%s: the device already has a rw-mount on %s."
-				"\n\t\tmultiple rw-mount is not supported."),
-			      progname, mc->m.mnt_dir);
-			goto failed;
-		}
+	if (mo->flags & MS_BIND)
+		return 0;
 
-		/* rw->rw remount */
-		if (check_remount_dir(mc, mi->mntdir) < 0)
-			goto failed;
-		
-		if (find_opt(mc->m.mnt_opts, gcpid_opt_fmt, &pid) >= 0)
-			mi->gcpid = pid;
-		mi->optstr = xstrdup(mc->m.mnt_opts); /* previous opts */
-		mi->type = RW2RW_REMOUNT;
-	} else if (mo->flags & MS_REMOUNT) { /* ro-remount */
-		if ((mc = find_rw_mount(mi->device)) == NULL)
-			goto out;
+	mc = find_rw_mount(mi->device);
+	if (mc == NULL)
+		return 0; /* no previous rw-mount */
+
+	/* get the value of previous pp option if exists */
+	prot_period = ULONG_MAX;
+	if (find_opt(mc->m.mnt_opts, pp_opt_fmt, &prot_period) >= 0)
+		mi->protperiod = prot_period;
+
+	mi->nogc = (find_opt(mc->m.mnt_opts, nogc_opt_fmt, NULL) >= 0);
+
+	switch (mo->flags & (MS_RDONLY | MS_REMOUNT)) {
+	case 0: /* overlapping rw-mount */
+		error(_("%s: the device already has a rw-mount on %s."
+			"\n\t\tmultiple rw-mount is not allowed."),
+		      progname, mc->m.mnt_dir);
+		goto failed;
+	case MS_RDONLY: /* ro-mount (a rw-mount exists) */
+		break;
+	case MS_REMOUNT | MS_RDONLY: /* rw->ro remount */
+	case MS_REMOUNT: /* rw->rw remount */
+		mi->type = (mo->flags & MS_RDONLY) ?
+			RW2RO_REMOUNT : RW2RW_REMOUNT;
 
-		/* rw->ro remount */
 		if (check_remount_dir(mc, mi->mntdir) < 0)
 			goto failed;
 		pid = 0;
 		if (find_opt(mc->m.mnt_opts, gcpid_opt_fmt, &pid) >= 0 &&
-		    !fake && stop_cleanerd(mi->device, (pid_t)pid) < 0) {
+		    nilfs_shutdown_cleanerd(mi->device, (pid_t)pid) < 0) {
 			error(_("%s: remount failed due to %s shutdown "
-				"failure"), progname, CLEANERD_NAME);
+				"failure"), progname, NILFS_CLEANERD_NAME);
 			goto failed;
 		}
 		mi->gcpid = pid;
 		mi->optstr = xstrdup(mc->m.mnt_opts); /* previous opts */
-		mi->type = RW2RO_REMOUNT;
+		break;
 	}
- out:
+
 	res = 0;
  failed:
 	return res;
@@ -386,10 +477,16 @@ prepare_mount(struct nilfs_mount_info *mi, const struct mount_options *mo)
 static int
 do_mount_one(struct nilfs_mount_info *mi, const struct mount_options *mo)
 {
-	int res, errsv;
+	int res, errsv, mtab_ok;
+	char *tmpexopts, *exopts;
+	pp_opt_t prot_period;
+
+	tmpexopts = change_opt(mo->extra_opts, pp_opt_fmt, &prot_period, "");
+	exopts = change_opt(tmpexopts, nogc_opt_fmt, NULL, "");
+	my_free(tmpexopts);
 
 	res = mount(mi->device, mi->mntdir, fstype, mo->flags & ~MS_NOSYS,
-		    mo->extra_opts);
+		    exopts);
 	if (!res)
 		goto out;
 
@@ -400,56 +497,70 @@ do_mount_one(struct nilfs_mount_info *mi, const struct mount_options *mo)
 		      progname, fstype);
 		break;
 	default:
-		error(_("%s: Error while mounting %s on %s: %s"), 
+		error(_("%s: Error while mounting %s on %s: %s"),
 		      progname, mi->device, mi->mntdir, strerror(errsv));
 		break;
 	}
-	if (mi->type != RW2RO_REMOUNT)
+	if (mi->type != RW2RO_REMOUNT && mi->type != RW2RW_REMOUNT)
 		goto out;
-	if (check_mtab()) {
+
+        mtab_ok = check_mtab();
+
+	/* Cleaner daemon was stopped and it needs to run */
+	/* because filesystem is still mounted */
+	if (!mi->nogc && mtab_ok) {
 		/* Restarting cleaner daemon */
-		if (start_cleanerd(mi->device, mi->mntdir, &mi->gcpid) == 0) {
+		if (nilfs_launch_cleanerd(mi->device, mi->mntdir,
+					  mi->protperiod, &mi->gcpid)) {
 			if (verbose)
 				printf(_("%s: restarted %s\n"),
-				       progname, CLEANERD_NAME);
+				       progname, NILFS_CLEANERD_NAME);
 			update_gcpid_opt(&mi->optstr, mi->gcpid);
 			update_mtab_entry(mi->device, mi->mntdir, fstype,
 					  mi->optstr, 0, 0, !mi->mounted);
 		} else {
 			error(_("%s: failed to restart %s"),
-			      progname, CLEANERD_NAME);
+			      progname, NILFS_CLEANERD_NAME);
 		}
 	} else
-		printf(_("%s not restarted\n"), CLEANERD_NAME);
+		printf(_("%s not restarted\n"), NILFS_CLEANERD_NAME);
  out:
+	my_free(exopts);
 	return res;
 }
 
 static void update_mount_state(struct nilfs_mount_info *mi,
 			       const struct mount_options *mo)
 {
-	pid_t pid = (mi->type == RW2RW_REMOUNT) ? mi->gcpid : 0;
+	pid_t pid = 0;
+	pp_opt_t pp = ULONG_MAX;
 	char *exopts;
 	int rungc;
 
-	rungc = !(mo->flags & MS_RDONLY) &&
-		(mi->type != RW2RW_REMOUNT || mi->gcpid == 0);
+	rungc = (find_opt(mo->extra_opts, nogc_opt_fmt, NULL) < 0) && !(mo->flags & MS_RDONLY) && !(mo->flags & MS_BIND);
+
 	if (!check_mtab()) {
 		if (rungc)
-			printf(_("%s not started\n"), CLEANERD_NAME);
+			printf(_("%s not started\n"), NILFS_CLEANERD_NAME);
 		return;
 	}
+
 	if (rungc) {
-		if (start_cleanerd(mi->device, mi->mntdir, &pid) < 0)
-			error(_("%s aborted"), CLEANERD_NAME);
+		if (find_opt(mo->extra_opts, pp_opt_fmt, &pp) < 0)
+			pp = mi->protperiod;
+		if (nilfs_launch_cleanerd(mi->device, mi->mntdir, pp,
+					  &pid) < 0)
+			error(_("%s aborted"), NILFS_CLEANERD_NAME);
 		else if (verbose)
-			printf(_("%s: started %s\n"), progname, CLEANERD_NAME);
+			printf(_("%s: started %s\n"), progname,
+			       NILFS_CLEANERD_NAME);
 	}
+
 	my_free(mi->optstr);
-	exopts = fix_extra_opts_string(mo->extra_opts, pid);
+	exopts = fix_extra_opts_string(mo->extra_opts, pid, pp);
 	mi->optstr = fix_opts_string(((mo->flags & ~MS_NOMTAB) | MS_NETDEV),
 				     exopts, NULL);
-		
+
 	update_mtab_entry(mi->device, mi->mntdir, fstype, mi->optstr, 0, 0,
 			  !mi->mounted);
 	my_free(exopts);
@@ -481,9 +592,7 @@ int main(int argc, char *argv[])
 {
 	struct mount_options *opts = &options;
 	char *device, *mntdir;
-	int res = 0; 
-
-	parse_options(argc, argv, opts);
+	int res = 0;
 
 	if (argc > 0) {
 		char *cp = strrchr(argv[0], '/');
@@ -491,6 +600,10 @@ int main(int argc, char *argv[])
 		progname = (cp ? cp + 1 : argv[0]);
 	}
 
+	nilfs_cleaner_logger = nilfs_mount_logger;
+
+	parse_options(argc, argv, opts);
+
 	umask(022);
 
 	if (optind >= argc || !argv[optind])
@@ -515,7 +628,7 @@ int main(int argc, char *argv[])
 		die(EX_USAGE, _("%s: no %s found - aborting"), progname,
 		    MOUNTED);
 
-	if (opts->flags & MS_RDONLY) {
+	if (!(opts->flags & MS_RDONLY) && !(opts->flags & MS_BIND)) {
 		res = device_is_readonly(device, &devro);
 		if (res)
 			die(EX_USAGE, _("%s: device %s not accessible: %s"),
@@ -537,7 +650,3 @@ int main(int argc, char *argv[])
 	my_free(opts->extra_opts);
 	return res;
 }
-
-/* Local Variables:		*/
-/* eval: (c-set-style "linux")	*/
-/* End:				*/
